unit uThreads;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils;

Type
  PMonitorComand=^TMonitorCommand;
  TTestThread=class;
  TBooleanArray=Array of Boolean;
  TStringArray=Array of String;

  TMCMDKey=(mcNone,mcSetCount,mcStart,mcStop,mcDestroy);
  TMStatus=(msNone,msIdle,msRunning,msStop,msDestroy);
  TMCMDKeyHistory=set of TMCMDKey;

  TMonitorCommand=record
    Key                          : TMCMDKey;
    Value                        : Int64;
  end;

  TMonitorThread=class(TThread)
  private
    FHistory                     : TMCMDKeyHistory;
    FStatus                      : TMStatus;
    FTerminated                  : TBooleanArray;
    FComplete                    : TBooleanArray;
    FComputing                   : TBooleanArray;
    FWatchList                   : TList;
    FLock                        : TRTLCriticalSection;
    FSleepP                      : PRTLEvent;
    FCmdQueue                    : TList;
  private
    procedure ResetCompletes;
    procedure ResetComputes;
    procedure ResetTerminated;
    function  AreAllComplete:boolean;
    function  AllAreTerminated:boolean;
    procedure UnWatch(Thread:TTestThread);
    procedure Watch(Thread:TTestThread);
    procedure ProcessCmdQueue;
    procedure ClearQueue;
    procedure SetThreadCount(Value:integer);
    function  GetThreadCount:integer;
    function  GetComputingCount:integer;
    function  GetComputingCount2:integer;
  protected
    procedure Execute; override; overload;
  public
    procedure Execute(Command:TMonitorCommand); overload;
  public
    constructor Create(); reIntroduce;
    destructor Destroy; override;
  public
    property Threads:Integer read GetThreadCount;
    property Computing:Integer read GetComputingCount;
    property Computing2:Integer read GetComputingCount2;
    property Status:TMStatus read FStatus;
  end;

  TTestThread=class(TThread)
  private
    FComputing                   : Boolean;
    FIndex                       : integer;
    FMonitorThread               : TMonitorThread;
    FSleepP                      : PRTLEvent;
  protected
    procedure Execute; override;
  public
    constructor Create(aMonitorThread:TMonitorThread); reIntroduce;
    destructor Destroy; override;
  end;

  procedure Init(Var Item:TMonitorCommand); overload;
  procedure Done(Var Item:TMonitorCommand); overload;
  procedure Copy(Var Source,Destination:TMonitorCommand); overload;

  function Command(aKey:TMCMDKey; aValue:Int64):TMonitorCommand;
  procedure Empty(Var Item:TStringArray);
var
  Manager:TMonitorThread;
const
  FMT_STATUS:Array[TMStatus] of string = ('sNone','msIdle','msRunning','msStop','msDestroy');

implementation

Const
  MAX_ITERATION:Cardinal=500000000;
  SUSPENDED:boolean=true;
  RUNNING:boolean=false;
  MONITOR_TIME_SLICE_YIELD_MS:cardinal=300;

  WORKER_TILE_SLICE_YIELD_MS:Array[boolean] of cardinal=(5000,200);


constructor TTestThread.Create(aMonitorThread:TMonitorThread);
begin
  FComputing:=false;
  FIndex:=-1;
  FSleepP:=RTLEventCreate;
  FreeOnTerminate:=True;
  Inherited Create(RUNNING);
  FMonitorThread:=aMonitorThread;
  FMonitorThread.Watch(self);
end;

destructor TTestThread.Destroy;
begin
  FMonitorThread.UnWatch(Self);
  RTLEventDestroy(FSleepP);
  FMonitorThread.FTerminated[FIndex]:=true;
  Inherited Destroy;
end;

function GenerateRandomString:string;
var
  iLcv:integer;
begin
  SetLength(Result,Random(9)+1);
  for iLcv:=1 to Length(Result) do
    Result[iLcv]:=Char(Random(200)+1);
end;

procedure TTestThread.Execute;
var
  iDataLcv,iLoop,iLcv:integer;
  iYield:cardinal;

  sLogData:string;
  saLogData:TStringArray;

begin
  Priority:=tpIdle;
  While (not Terminated ) and (FMonitorThread.Status<>msDestroy) do begin
    // Do some log work here.
    iYield:=WORKER_TILE_SLICE_YIELD_MS[FMonitorThread.FStatus=msRunning];
    if FMonitorThread.FStatus=msRunning then begin
      // Does execution here happen out of expected order?
      // Order execution w/r/t memory assigments are affected how?
      if not FMonitorThread.FComplete[FIndex] then begin
        FComputing:=true;
        FMonitorThread.FComputing[FIndex]:=FComputing;
        Try
          iLoop:=MAX_ITERATION;
          for iLcv:=0 to iLoop do begin
            if Terminated or (FMonitorThread.FStatus<>msRunning) then break;
            SetLength(saLogData,Random(25)+1); // << allocate 1 log entry with no more than 25 fields
            Try
              for iDataLcv:=0 to High(saLogData) do
                saLogData[iDataLcv]:=GenerateRandomString;
            finally
              Empty(saLogData);
            end;
            if FMonitorThread.FComplete[FIndex] then
              FMonitorThread.FComplete[FIndex]:=True; // Break Point Proof of premature completion
            if FComputing=false then
              FComputing:=true; // Break point proof of premature computing
            if (iLcv mod 1000=0) then RTLeventWaitFor(FSleepP,iYield);
          end;
          FMonitorThread.FComplete[FIndex]:=True;
          If FMonitorThread.FComputing[FIndex]<>FComputing then
            FComputing:=FMonitorThread.FComputing[FIndex];    // Proof of out of order executing
        finally
          FComputing:=false;
          FMonitorThread.FComputing[FIndex]:=FComputing;
        end;
      end;
    end;
    RTLeventWaitFor(FSleepP,iYield);
    iYield:=0;
  end;
end;

constructor TMonitorThread.Create();
begin
  FHistory:=[];
  FStatus:=msNone;
  FreeOnTerminate:=True;
  FSleepP:=RTLEventCreate;
  FWatchList:=TList.Create;
  FCmdQueue:=TList.Create;
  InitCriticalSection(FLock);
  Inherited Create(RUNNING);
end;

destructor TMonitorThread.Destroy;
begin
  ClearQueue;

  FreeAndNil(FWatchList);
  FreeAndNil(FCmdQueue);
  DoneCriticalSection(FLock);
  RTLEventDestroy(FSleepP);

  Finalize(FComplete);
  Finalize(FComputing);
  Finalize(FTerminated);
  Inherited Destroy;
end;

function TMonitorThread.AreAllComplete:boolean;
var
  iCount,iLcv:integer;
begin
  Result:=true; iLcv:=0; iCount:=Length(FComplete);
  While (Result) and (iLcv<iCount) do begin
    Result:=FComplete[iLcv];
    Inc(iLcv);
  end;
end;

function  TMonitorThread.AllAreTerminated:boolean;
var
  iCount,iLcv:integer;
  bExit:boolean;
begin
  Result:=True; bExit:=false; iLcv:=0; iCount:=Length(FTerminated);
  While (bExit=false) and (iLcv<iCount) do begin
    bExit:=(FTerminated[iLcv]=false);
    Inc(iLcv);
  end;
  Result:=(bExit=false);
end;

procedure TMonitorThread.ResetCompletes;
var
  iLcv:integer;
begin
  EnterCriticalSection(FLock);
  Try
    for iLcv:=0 to High(FComplete) do
      FComplete[iLcv]:=false;
  finally
    LeaveCriticalSection(FLock);
  end;
end;

procedure TMonitorThread.ResetComputes;
var
  iLcv:integer;
begin
  EnterCriticalSection(FLock);
  Try
    for iLcv:=0 to High(FComputing) do
      FComputing[iLcv]:=false;
  finally
    LeaveCriticalSection(FLock);
  end;
end;

procedure  TMonitorThread.ResetTerminated;
var
  iLcv:integer;
begin
  EnterCriticalSection(FLock);
  Try
    for iLcv:=0 to High(FComputing) do
      FTerminated[iLcv]:=false;
  finally
    LeaveCriticalSection(FLock);
  end;
end;

procedure  TMonitorThread.Execute;
begin
  Priority:=tpNormal;
  While Not Terminated do begin
    ProcessCmdQueue;
    Case FStatus of
      msRunning : If AreAllComplete then FStatus:=msIdle;
      msStop    : If AreAllComplete then FStatus:=msIdle;
      msDestroy : If AllAreTerminated then Terminate;
    end;
    RTLeventWaitFor(FSleepP,MONITOR_TIME_SLICE_YIELD_MS);
  end;
end;

procedure TMonitorThread.Execute(Command:TMonitorCommand);
var
  cmdP:PMonitorComand;
begin
  New(cmdP);
  Init(cmdP^);
  Copy(Command,cmdP^);
  EnterCriticalSection(FLock);
  Try
    FCmdQueue.Add(cmdP);
  finally
    LeaveCriticalSection(FLock);
  end;
end;

procedure TMonitorThread.UnWatch(Thread:TTestThread);
var
  iLcv:integer;
begin
  EnterCriticalSection(FLock);
  Try
    FComplete[Thread.FIndex]:=true;
    FComputing[Thread.FIndex]:=false;
    FWatchList.Items[Thread.FIndex]:=nil;
  finally
    LeaveCriticalSection(FLock);
  end;
end;

procedure TMonitorThread.Watch(Thread:TTestThread);
begin
  if (Thread.FIndex=-1) or (FWatchList[Thread.FIndex]<>nil) then begin
    Thread.FIndex:=FWatchList.Add(Thread);
  end else begin
    FWatchList[Thread.FIndex]:=Thread;
  end;
end;

function  TMonitorThread.GetThreadCount:integer;
begin
  EnterCriticalSection(FLock);
  Try
    Result:=FWatchList.Count;
  finally
    LeaveCriticalSection(FLock);
  end;
end;

function  TMonitorThread.GetComputingCount:integer;
var
  iCount:integer;
  iLcv:integer;
  tdWorker:TTestThread;
begin
  Result:=0;
  EnterCriticalSection(FLock);
  Try
    iCount:=FWatchList.Count;
    for iLcv:=0 to iCount-1 do begin
      tdWorker:=TTestThread(FWatchList[iLcv]);
      if (tdWorker<>nil) and (tdWorker.FComputing) then
        Inc(Result);
    end;
  finally
    LeaveCriticalSection(FLock);
  end;
end;

function  TMonitorThread.GetComputingCount2:integer;
var
  iLcv:integer;
begin
  Result:=0;
  For iLcv:=0 to High(FComputing) do
    if FComputing[iLcv] then Inc(Result);
end;

procedure TMonitorThread.SetThreadCount(Value:integer);
var
  tdWorker:TTestThread;
begin
  EnterCriticalSection(FLock);
  Try
    While (FWatchList.Count<Value) do
      tdWorker:=TTestThread.Create(self);
    SetLength(FComplete,FWatchList.Count);
    SetLength(FComputing,FWatchList.Count);
    SetLength(FTerminated,FWatchList.Count);
    ResetTerminated;
  finally
    LeaveCriticalSection(FLock);
  end;
end;

procedure TMonitorThread.ClearQueue;
var
  iCount,iLcv:integer;
  cmdP:PMonitorComand;
begin
  EnterCriticalSection(FLock);
  Try
    iCount:=FCmdQueue.Count;
    While (iCount>0) do begin
      cmdP:=FCmdQueue.Items[0];
      if cmdP<>nil then begin
        Done(cmdP^);
        Dispose(cmdP);
      end;
      FCmdQueue.Delete(0);
      Dec(iCount);
    end;
  finally
    LeaveCriticalSection(FLock);
  end;
end;

procedure TMonitorThread.ProcessCmdQueue;
var
  iLcv:integer;
  cmdP:PMonitorComand;

  procedure ProcessNone;
  begin

  end;

  procedure ProcessSetCount;
  begin
    if not (mcSetCount in FHistory) then begin
      // Safe to enforce destired Count
      Include(FHistory,mcSetCount);
      SetThreadCount(cmdP^.Value);
    end;
  end;

  procedure ProcessStart;
  begin
    if (mcSetCount in FHistory) and (FStatus<>msRunning) then begin
      // Safe to startup system
      ResetCompletes;
      ResetComputes;
      FStatus:=msRunning;
      Include(FHistory,mcStart);
    end;

  end;

  procedure ProcessStop;
  begin
    if (mcSetCount in FHistory) and (mcStart in FHistory) then begin
      // Safe to stop system
      FStatus:=msStop;
      Include(FHistory,mcStop);
    end;
  end;

  procedure ProcessDestroy;
  begin
    FStatus:=msDestroy;
    Include(FHistory,mcDestroy);
  end;

begin
  EnterCriticalSection(FLock);
  Try
    For iLcv:=0 to FCmdQueue.Count-1 do begin
      cmdP:=FCmdQueue.Items[iLcv];
      if cmdP<>nil then begin
        case cmdP^.Key of
          mcNone     : ProcessNone;
          mcSetCount : ProcessSetCount;
          mcStart    : ProcessStart;
          mcStop     : ProcessStop;
          mcDestroy  : ProcessDestroy;
        end;
        Done(cmdP^);
        Dispose(cmdP);
        FCmdQueue.Items[iLcv]:=nil;
      end;
    end;
    FCmdQueue.Pack;
  Finally
    LeaveCriticalSection(FLock);
  end;
end;

procedure Init(Var Item:TMonitorCommand);
begin
  Item.Key:=mcNone;
  Item.Value:=0;
end;

procedure Done(Var Item:TMonitorCommand);
begin
  Finalize(Item);
end;

procedure Copy(Var Source,Destination:TMonitorCommand);
begin
  Destination.Value:=Source.Value;
  Destination.Key:=Source.Key;
end;

function Command(aKey:TMCMDKey; aValue:Int64):TMonitorCommand;
begin
  Result.Key:=aKey;
  Result.Value:=aValue;
end;

procedure Empty(Var Item:TStringArray);
var
  iLcv:integer;
begin
  for iLcv:=0 to High(Item) do
    Finalize(Item[iLcv]);
  SetLength(Item,0);
end;

end.

